! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
  module psf

  use param
  use dal
  use cal
  use emldata 
  use edetect_io_ml
  implicit none

  private
  public :: bpsf, psf4 !!!, psf5, psf6

contains

!!!  function psfoff (ene, eps, x, sig, ierr) result (result_psfoff) 
  function psfoff (eps, x, sig) result (result_psfoff) 

!  calculates XMM off-axis PSF (normalized surface brightness)
!
!  author : HEB        date: 15-NOV-1996 09:21 
!  modifi : GRH        date: 15-NOV-1996 20:07  off-axis from ABRIXAS 
!  modifi : GRH        date: 19-NOV-1996 09:50  XMM simulations included
!
!  The XMM QM PSF has been fit by Mike Watson with 4 Gaussians  
!  GRH has normalized the PSF components and added a mirror-blur term in
!  quadrature as a function of off axis angle, derived from XMM 
!  simulations performed by P.Friedrich using the ABRIXAS sim code.
!
!   call_var.          type I/O description
!  ENE                 R   I   Energy [keV]
!  EPS                  R   I      off-axis angle [arc min]
!  X                   R   I   angle from target position [arcsec]
!
!  IERR                I   O   = 0 no error 
!                              = 1 energy outside bounds 0.07-3 keV
!                              = 3 
!
!     OUTPUT:
!             PSFOFF      surface brightness of point spread function,
!                         normalized, so that  Integral 2*PI*r*dr*f
!                         from 0 to infinity = 1 [1/arcsec2]

    real          result_psfoff
    real          eps, x, sig, fac   !!!  ene, 
!!!    integer       ierr
    real          norm(4),sigma(4),ssum,offsg,sig2,rnorm,sigtel
    integer       i

! MGW data norm  /0.75337,0.14463,0.79143e-2,0.75355e-3/,
! following are the norms after integrating each component to 1.00
    data norm  /0.4169,0.3470,0.1279,0.1082/, &
         sigma /4.1450,8.6218,22.372,66.698/, &
         offsg /0.0324/

    ssum=0.
! off-axis component of telescope    
    sigtel=offsg*min(eps,18.0)**2
    fac=sigtel*sigtel+sig*sig
    do i=1,4
      sig2=0.5/(sigma(i)**2+fac)
      rnorm=norm(i)*sig2*overpi
      ssum=ssum+rnorm*exp(-sig2*(x)**2)
    end do

    result_psfoff = ssum

  end function psfoff


!---------------------------------------------------------------------------------------

  real function bpsf (energy, iband, x1, y1, dx0, dy0, dsig, ecut,offax,azim,n) &
       result (result_bpsf)

!!! IT 2012/01
!!! bpsf ("background") is a new version of former psf1, psf2, psf3 functions,
!!! including the new re-sampling algorithm for the 2d ellbeta PSF.
!!! Removed lots of old comments and old code that are no longer valid.
!!! For the old versions, see "psf.f90oldalgorithm".

    real(kind=double), intent(in) ::  dx0, dy0, dsig
    real, intent(in) :: ecut
    real ::  x0, y0, sig
    real, dimension(max_ima), intent(in) :: energy
    real(kind=double), dimension(max_ins,max_fit)  :: offax,azim
    integer, intent(in) :: n
!    real  intent(in) :: ene
    integer, intent(in) :: iband, x1, y1
!!! IT 2012/01
!    integer :: j,k
    integer :: ninst, nima

!!! IT 2012/01
!!! sec2mm will be calculated for each instrument individually, using the CAL
!    real, parameter :: sec2mm = 7500.0*pi/(180.0*3600.0)
    real(kind=double), dimension(max_ima) :: sec2mm
    real(kind=double) :: E, theta, phi, val,  focalLength
    logical :: first = .true.
!!!    integer :: i,ix,iy,ixx,iyy,lx,ly,kx,ky,kk,np,np2,isig3,isig3sq
    integer :: i,ix,iy,lx,ly,kx,ky,kk,np,isig3,isig3sq
    real :: np2
    integer, dimension(max_ima) :: nx, ny
    type (PsfT) :: ppsf
    type(PsfBinSizeT), dimension(max_ima) :: bsiz
    type(PSFwindowT) :: window
    type(PsfValidityRangesT) :: vrange 
    real(kind=single), dimension(:,:), pointer :: pima
    real, dimension(:,:,:), allocatable :: sima
!!    real, dimension(:,:), allocatable :: pima2
!!    real, dimension(:,:,:), allocatable ::  psave
!!! IT 2012/07: make gengen allocatable
!    real, dimension(0:100,0:100) :: gengen
    real, dimension(:,:), allocatable :: gengen
!!! IT 2012/01
!    real :: x,y,dx,dy,ddx,ddy,xd0,yd0,xx,yy,x0old,y0old, &
!            ssum,winr,pfac,xyoff2,ecutold
    real :: x,y,dx,dy,ddx,ddy,xd0,yd0,xx,yy,x0old,y0old, &
            ssum,winr,pfacX,pfacY,xyoff2,ecutold
    real(kind=double)   :: pxmin,pxmax,pymin,pymax,dx1,dx2,dy1,dy2
    integer             :: pxmin0,pymin0,pxmax0,pymax0,pxmin1,pymin1
    integer astat
!!!---------------
    real :: sigold = -1.0, isig3old = -1.0
    integer :: sxlo,sxhi,sylo,syhi   !!!,nlo,nhi
!    integer :: cl1,cl2,cl3,cl0
!---JR
!    real :: testsum
!    integer :: ix1, iy1, ixnew
    save

    x0=sngl(dx0)
    y0=sngl(dy0)
    sig=sngl(dsig)

    xyoff2=(120.0/pixim)**2    

!!! IT 2012/07: make gengen allocatable with size isig3
!!! i.e. scale it with source extent and image binning (cf. SPR-6747)
! initialize for extend calculation to 10 sigma     (copied from below)
      isig3 = nint(min(10.*sig,ecut))
      if (isig3/=isig3old .or. first) then
         if (.not. first) deallocate(gengen)
         allocate(gengen(0:isig3,0:isig3), stat=astat)
         if (astat /= 0) call error("notEnoughMemory", "Not enough memory for array 'gengen'")
         gengen(:,:) = 0.
         gengen(0,0) = 1.0
         isig3old = isig3
      endif

    if (first) then
!      count_model = 0
      allocate (sima(0:650,0:650,nim), stat=astat) !!  to 450 - JR
      if (astat /= 0) call error("notEnoughMemory", "Not enough memory for array 'sima'")
      sima = 0.0
!!      nlo=30
!!      nhi=30
!!      allocate (psave(nim,nlo:nhi,nlo:nhi))
!      gengen(0,0) = 1.0
    end if 

    if ((x0-x0old)**2+(y0-y0old)**2>xyoff2.or.ecut/=ecutold.or.first) then

!!       psfnorm=1.0

     do ninst=1,num_instr

!!! set PN ; use PN also for "unknown" instruments (-> from former emlfit code)

!!! IT 2014/08
!!! - declutter code by using select-case syntax
!!! - use of extpsf and ACCURACY_EXTENDED are deprecated
!      if (extpsf) then
!        call CAL_setState1(instrument=EPN, AccuracyLevel=ACCURACY_EXTENDED) 

      ! set instrument & accuracy level
      select case (expip(ninst))
      case (2)                        !!! MOS 1
         select case (psfmod)
         case ("medium")
            call CAL_setState1(instrument=EMOS1,AccuracyLevel=ACCURACY_MEDIUM)
         case ("slew")
            call CAL_setState1(instrument=EMOS1,AccuracyLevel=ACCURACY_SLEW)
         case default    !!! "ellbeta"
            call CAL_setState1(instrument=EMOS1,AccuracyLevel=ACCURACY_ELLBETA)
         end select
      case (3)                        !!! MOS 2
         select case (psfmod)
         case ("medium")
            call CAL_setState1(instrument=EMOS2,AccuracyLevel=ACCURACY_MEDIUM)
         case ("slew")
            call CAL_setState1(instrument=EMOS2,AccuracyLevel=ACCURACY_SLEW)
         case default    !!! "ellbeta"
            call CAL_setState1(instrument=EMOS2,AccuracyLevel=ACCURACY_ELLBETA)
         end select
      case default ! (1)              !!! PN (& unknown)
         select case (psfmod)
         case ("medium")
            call CAL_setState1(instrument=EPN,AccuracyLevel=ACCURACY_MEDIUM)
         case ("slew")
            call CAL_setState1(instrument=EPN,AccuracyLevel=ACCURACY_SLEW)
         case default    !!! "ellbeta"
            call CAL_setState1(instrument=EPN,AccuracyLevel=ACCURACY_ELLBETA)
         end select
      end select

!!!-----------------------------------------------------------------------------

      theta=offax(ninst,n)
!!      phi=azim(ninst,n)-pa_pnt(ninst)*pi180+pi2
!!      if (phi .gt. twopi) phi=phi-twopi
!!      if (phi .lt. -twopi) phi=phi+twopi
      phi=mod(azim(ninst,n)-pa_pnt(ninst)*pi180+pi2, twopi)

      xd0=x0-xcen(ninst)
      yd0=y0-ycen(ninst)

      vrange = CAL_PsfValidityRanges()
      theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)

      np = max(0,nint((pixim-1.0)*0.5))
      np2= 1./(2*np+1)**2 

      winr = int(ecut)+2.0
!-------------------------------------
!-- JR -------------------------------
      if (psfmod .eq. "medium") then
         winr=  min(winr*pixim,100.0)      !!! *sec2mm     -> later
      else
         winr=  min(winr*pixim,400.0)      !!! *sec2mm     -> later
      endif
!-------------------------------------
!-------------------------------------

      do nima=1,num_image(ninst)
        i=ip(ninst,nima) 

!!! IT 2012/01
!!! get focal length of the telescope from the CAL (for each instrument)
!!! calculate conversion factors "sec2mm" (arcsec to mm)
        if (first) then
           !!! IT 2012/03: new version only for ellbeta to be consistent
           !!! with old medium-accuracy positions
           if (psfmod .eq. "medium") then
              sec2mm(i) = 75d0*pi180/36d0    !7500.0*pi/(180.0*3600.0)
           else
              call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLength)
              sec2mm(i) = focalLength*pi180/3600.d0
           endif
        endif

        window%halfX = winr * sec2mm(i)
        window%halfY = winr * sec2mm(i)

        E = energy(i)
        E = max(min(E,vrange%eMax-1.0),vrange%eMin+1.0)

! -- according to Richard ----------------------------------------
        ppsf = Cal_getPSF(E, theta, azim(ninst,n))
        call CAL_psfSetPosAngle(ppsf, pa_pnt(ninst)*pi180)


        bsiz(i) = CAL_psfBINSIZE(ppsf) 
        call CAL_psfImage (ppsf, window, pima)
                                                                             
!       fudge factor for EXTENDED PSF:
!!! IT 2012/01: psfnorm is 1.
!!!        pima=pima*psfnorm
        nx(i) = size(pima,1)
        ny(i) = size(pima,2)

!!!-----------------------------------------------------------------------------
!!! IT 2012/01
!!! new re-sampling algorithm for the ellbeta 2d PSF
!!! For details, see fpsf in fftpsf.f90 .
!!!

         if (psfmod .eq. "medium") then
!!!------> old version of re-sampling (1)
!!!        smoothing the medium accuracy PSF with a box of image pixel size
            ! smooth PSF with (2*np+1) sized box:
            do iy=np,ny(i)-1-np
               do ix=np,nx(i)-1-np
!!! IT 2012/01: faster
!            ssum=0.0
!            do ixx=ix-np,ix+np
!              do iyy=iy-np,iy+np
!                ssum=ssum+pima(ixx,iyy)
!              end do
!            end do
!            sima(i,ix,iy)=ssum/np2
                  sima(ix,iy,i) = sum(pima(ix-np:ix+np,iy-np:iy+np)) * np2
               end do
            end do
!!!<--------
         else

!!! CAUTION!!! pima indices start with 0, not with 1
            sima(0:nx(i)-1,0:ny(i)-1,i) = pima

         endif

        call CAL_releaseMemory(pima)
      end do
     end do


      ecutold=ecut
      first = .false.
!!      psave=-1.
    end if

!!!-----------------------------------------------------------------------------
!!!-----------------------------------------------------------------------------

!     count_model = count_model + 1

      x0old=x0
      y0old=y0 

! extend calculation to 10 sigma (06/11/05)
!!!      isig3 = nint(min(10.*sig,ecut))    ! copied to beginning of code
!!!      isig3sq = isig3*isig3

      if (sig == 0.0) then  
        gengen=0.0
        gengen(0,0)=1.0     
      endif
      if (sig/=sigold .and. sig > 0.0) then
        isig3sq = isig3*isig3
        if (extmod == "GAUSSIAN") then         !               Gaussian extent:
        do ly = 0,isig3
          y=ly
          do lx = 0,ly
            if ((lx*lx+ly*ly)<=isig3sq) then        
              x=lx
                if (lx/=ly) then
                  gengen(lx,ly) = gen(0.0,x-0.5,x+0.5,sig)* &
                                  gen(0.0,y-0.5,y+0.5,sig)
                else
                  gengen(lx,lx) = gen(0.0,x-0.5,x+0.5,sig)**2
                end if
            end if 
          end do
        end do
     else                                     !               Beta model:
        do ly = 0,isig3
          y=ly
          do lx = 0,ly
            if ((lx*lx+ly*ly)<=isig3sq) then        
              x=lx
              gengen(lx,ly)=king(x,y,sig) 
            end if 
          end do
        end do
      end if
   end if
      sigold=sig


!!!-----------------------------------------------------------------------------
!!! IT 2012/01
!!! new re-sampling algorithm for the ellbeta 2d PSF
!!! For details, see fpsf in fftpsf.f90 .
!!!

!      pfac=pixim*sec2mm/bsiz(iband)%binX
      pfacX=pixim*sec2mm(iband)/bsiz(iband)%binX
      pfacY=pixim*sec2mm(iband)/bsiz(iband)%binY

      if (psfmod .ne. "medium") then        !!! -> ELLBETA (2D) PSF

!!! lower left corner of the image pixel that contains the source centre,
!!! expressed in PSF coordinates
!!!   i.e. center of PSF (nx/2.,ny/2.) - shift - half an image pixel
         xx = dble(nx(iband)*0.5) + (x1-x0 - 0.5d0 )*pfacX
         yy = dble(ny(iband)*0.5) + (y1-y0 - 0.5d0 )*pfacY

!!! area in which the PSF shall be re-sampled, expressed in image coordinates
         sxlo = max(-isig3,int((1-xx)/pfacX))           
         sxhi = min(isig3,int((nx(iband)-2-xx)/pfacX)+1)
         sylo = max(-isig3,int((1-yy)/pfacY))           
         syhi = min(isig3,int((ny(iband)-2-yy)/pfacY)+1)
!if (isig3.ne.0) write(*,*) "using ",isig3,int((1-xx)/pfacX),int((nx(iband)-2-xx)/pfacX)+1,int((ny(iband)-2-yy)/pfacY)+1
!if (syhi.lt.0 .or. syhi.lt.0) write(*,*) "WARNING -- ",sxlo,sxhi,sylo,syhi

!!! re-sampling
         ssum = 0.
         do ly = sylo,syhi   ! -isig3,isig3 
            !! loop variables describing image pixel edges in PSF coordinates
            pymin = yy+ly*pfacY                                !! formerly: y
            pymin0 = int(pymin)
            pymin1 = pymin0 + 1
            pymax = pymin + pfacY
            pymax0 = int(pymax)
            !! cut to usable area
!omitted!
            !! coefficients used in the re-binning term
            dy1 = pymin1 - pymin
            dy2 = pymax - pymax0

            do lx =  sxlo,sxhi ! -isig3,isig3 
               !! loop variables describing image pixel edges in PSF coordinates
               pxmin = xx+lx*pfacX                                !! formerly: x
               pxmin0 = int(pxmin)
               pxmin1 = pxmin0 + 1
               pxmax = pxmin + pfacX
               pxmax0 = int(pxmax)
               !! cut to usable area
!omitted!
               !! coefficients used in the re-binning term
               dx1 = pxmin1 - pxmin
               dx2 = pxmax - pxmax0

               !! only, if complete pixels in x and / or y are covered:
               val=0.0
               if (pxmax0 > pxmin1 ) then
                  !! edges: complete in x, fraction in y
                  val = val +                                            &
                       dy1 * sum(sima(pxmin1:pxmax0-1, pymin0, iband)) + &
                       dy2 * sum(sima(pxmin1:pxmax0-1, pymax0, iband))
               endif
               if (pymax0 > pymin1 ) then
                  !! edges: complete in y, fraction in x
                  val = val +                                            &
                       dx1 * sum(sima(pxmin0, pymin1:pymax0-1, iband)) + &
                       dx2 * sum(sima(pxmax0, pymin1:pymax0-1, iband))
                  if (pxmax0 > pxmin1 ) then
                     !! whole (PSF) pixels
                     val = val +                                         &
                          sum(sima(pxmin1:pxmax0-1, pymin1:pymax0-1, iband))
                  endif
               endif

               !! always: corners, subpixel contributions
               val = val +                                               &
                    dx1 * dy1 * sima(pxmin0, pymin0, iband) +            &
                    dx1 * dy2 * sima(pxmin0, pymax0, iband) +            &
                    dx2 * dy1 * sima(pxmax0, pymin0, iband) +            &
                    dx2 * dy2 * sima(pxmax0, pymax0, iband)

!!! identical to old version
               kx = abs(lx)
               ky = abs(ly)
               if (kx>ky) then
                  kk=kx
                  kx=ky
                  ky=kk
               end if
               ssum = ssum + val * gengen (kx,ky)

            end do
         end do

!!! IT 2012/02 with the new re-binning, the normalization factor is 1
!!!        ssum = ssum * pixim*pixim

      else                    !!! -> MEDIUM ACCURACY PSF

!!!------> old version of re-sampling (2)
!!!        interpolating the medium accuracy PSF from CAL px to image px
         xx = ((x1-x0)*pixim*sec2mm(iband)+window%halfX)/bsiz(iband)%binX-0.5
         yy = ((y1-y0)*pixim*sec2mm(iband)+window%halfY)/bsiz(iband)%binY-0.5   

         sxlo = max(-isig3,int((1-xx)/pfacX))
         sxhi = min(isig3,int((nx(iband)-2-xx)/pfacX)+1)
         sylo = max(-isig3,int((1-yy)/pfacY))
         syhi = min(isig3,int((ny(iband)-2-yy)/pfacY)+1)

         ssum = 0.
         do ly = sylo,syhi   ! -isig3,isig3 
            y=yy+ly*pfacY 
            iy = min(max(int(y),np),ny(iband)-2-np)
            dy=max(y-iy,0.)
            ddy=max(1.-dy,0.)
            do lx =  sxlo,sxhi ! -isig3,isig3 
               x=xx+lx*pfacX
               ix = min(max(int(x),np),nx(iband)-2-np)
               dx=max(x-ix,0.)
               ddx=max(1.-dx,0.)

               val = ddx*ddy*sima(ix  ,iy  ,iband)   + &
                    ddx* dy*sima(ix  ,iy+1,iband)   + &
                    dx*ddy*sima(ix+1,iy  ,iband)   + &
                    dx* dy*sima(ix+1,iy+1,iband)
        
               kx = abs(lx)
               ky = abs(ly)
               if (kx>ky) then
                  kk=kx
                  kx=ky
                  ky=kk
               end if
               ssum = ssum + val * gengen (kx,ky)

            end do
         end do
!!!<------

        ssum = ssum * sec2mm(iband)*sec2mm(iband)*pixim*pixim

      endif


!      call CAL_releaseMemory(gengen)

!!!-----------------------------------------------------------------------------

101  format(377(1pe9.3,1x))

      result_bpsf = ssum



!  call system_clock(cl1,cl2,cl3)
!  write(*,*) "psf time:", float(cl1-cl0)/1000.

  end function bpsf


!---------------------------------------------------------------------------------------

!!!  real function psf4 (energy, iband, x1, y1, dx0, dy0, xcen, ycen, dsig, ecut,theta) &
!!!       result (result_psf4)
  real function psf4 (x1, y1, dx0, dy0, xcen, ycen, dsig, theta) &
       result (result_psf4)


    real(kind=double), intent(in) ::  dx0, dy0, dsig
    real, intent(in) :: xcen,ycen   !!! ecut,
    real ::  x0, y0, sig
    real(kind=double), intent(in) :: theta
!!!    real, dimension(max_ima), intent(in) :: energy
!!!    integer, intent(in) :: iband
    real ::  x1, y1
!!!    integer :: ierr

    real    :: x, eps, xsig, xd0, yd0, rc2, r2
       
  
    x0=sngl(dx0)
    y0=sngl(dy0)
    sig=sngl(dsig)



    xd0=x0-xcen
    yd0=y0-ycen
      
    if (extmod == "GAUSSIAN" .or. sig .lt. 2.0) then 
      x = sqrt ((x1-x0)**2+(y1-y0)**2) * pixim
      xsig = sig * pixim
!    eps = sqrt(xd0**2+yd0**2)*pixim/60.
      eps = theta/60.
!!!      result_psf4 = psfoff (1.0, eps, x, xsig, ierr) * pixim*pixim
      result_psf4 = psfoff (eps, x, xsig) * pixim*pixim
    else
      !beta model:
      rc2=sig*sig
      r2=(x1-x0)*(x1-x0)+(y1-y0)*(y1-y0)
      result_psf4=(1+r2/rc2)**(-1.5)/(twopi*rc2)
    endif

  end function psf4
!---------------------------------------------------------------------------------------
!!! OBSOLETE
!!!
!!!  real function psf5 (ene, offax, posang, x1, y1, x0, y0, sig)
!!!
!!!    real, intent(in) :: ene, offax, posang, x0, y0, x1, y1, sig
!!!    psf5 = 0.
!!!
!!!  end function psf5
!!!
!!!!---------------------------------------------------------------------------------------
!!!  real function psf6 (ene,offax,posang,x1,y1,x0,y0,sig)
!!!
!!!    real, intent(in) :: ene, offax, posang, x0, y0, x1, y1, sig
!!!    psf6 = 0.
!!!
!!!  end function psf6
!!!
!---------------------------------------------------------------------------------------

  real function gen(x0,x1,x2,sig)

    real x0,x1,x2,sig,fsig
                                                                       
    gen=0.
    fsig=sig                                                           
    if (fsig.le.0.00001) fsig=0.00001
    fsig = 1.0/(1.414214*fsig)
    gen = (errf((x2-x0)*fsig)-errf((x1-x0)*fsig))*0.5

  end function gen     

!---------------------------------------------------------------------------------------

  real function errf (xx)

    real t,x,xx

    x=abs(xx)                                                        
    errf=1.0
                                                         
    if (x<=5.0) then                                            
       t=1.0/(1.0+0.47047*x)                                            
       errf=1.0-(0.3480242-(0.0958798-0.7478556*t)*t)*t*exp(-(x*x))
     end if   
                                                                    
     if (xx<0.0) errf=-errf                                        

   end function errf     

!---------------------------------------------------------------------------------------
   real function king(x,y,rcore)

!  returns value of the King profile (1+r**2/rcore**2)**(-1.5) integrated over the pixel with center (x,y)
!  normalized to 1  
!  the integration is approximated by the radial integral between r1 and r2, r1=r_pix-0.5, r2=r_pix+0.5

     real x,y,rcore
     real r1,r2,r1sq,r2sq,rc2,r
     real norm   !!! ,kingint,area


     rc2=max(1E-2,rcore*rcore)
  
     r=sqrt(x*x+y*y)
     r1=max(0.,r-0.5)
     r2=r+0.5    
     r1sq=r1*r1
     r2sq=r2*r2 
!-----------------------------------------------------
!     kingint=twopi*(rc2/(1+r1sq/rc2)-rc2/(1+r2sq/rc2))
!     area=pi*(r2sq-r1sq)
!     norm=twopi*rc2
!     king=kingint/area/norm 
!----------------------------------------------------

!    Correct integration of king model (26/06/05 GL):
     norm=twopi*rc2
     king=(r2/sqrt(1.+r2sq/rc2)-(r1/sqrt(1.+r1sq/rc2)))/(r2-r1)/norm
 

   end function king

!---------------------------------------------------------------------------------------

end module psf





